==> kaya_0.2.5.orig.tar.gz: kaya-0.2.5/rts/Heap.h <==
#ifndef _HEAP_H // -*-C++-*-
#define _HEAP_H

/**
    Kaya run-time system
    Copyright (C) 2004, 2005 Edwin Brady

    This file is distributed under the terms of the GNU Lesser General
    Public Licence. See COPYING for licence.
*/

using namespace std;

#ifndef WIN32
#define SWPRINTF swprintf
#else
#define SWPRINTF snwprintf
#endif
#ifndef NOCHECK
#define GC_DEBUG
#endif
#ifndef WIN32
#include <pthread.h>
#define GC_PTHREADS
#else
#define GC_WIN32_THREADS
#endif
#include <gc/gc_cpp.h>

//#include <gc/gc_allocator.h>
#include <vector>
#include <cstdlib>
#include "sizes.h"
#include "ValueFuns.h"

class VMState;

typedef void(*func)(VMState*);

class Value;
class String;
class Array;
class Exception;
class Union;
class Closure;

// I really shouldn't call this Real! Float/Double better.
class Real : public gc
{
public:
    Real(double n):number(n) {}
    double number;
};

// cached values
extern Value *zero, *one, *minusone;

extern "C" {
  wchar_t* emptyString();
}


#define U_TAG(x) ((x->tag_arity) >> 16)
#define U_ARITY(x) ((x->tag_arity) & 65535)
#define THIS_TAG (tag_arity >> 16)
#define THIS_ARITY (tag_arity & 65535)

class Union : public gc_cleanup
{
public:
    // If getargs is true, get arguments from the stack in vm.
    Union(VMState* vm, int tag,int arity, bool getargs = true);
    ~Union() { if (args != NULL) { GC_FREE(args); } }

    bool eq(Union* x, map<Value*,Value*>& done);
    int cmp(Union* x);
    int memusage();
    Union* copy(map<Value*,Value*>& done);

    int tag_arity; // tag in top 16 bits, arity in bottom
    Value** args;
};

struct CallStackEntry : public gc {
    wchar_t* fn_name;
    wchar_t* file;
    int line;
    wchar_t* call_file;
    int call_line;
};

class Exception : public gc
{
public:
    Exception(VMState* vm);
    Exception(VMState* vm, char* tag, int arity);
    void show();
    bool eq(Exception* x);
    int cmp(Exception* x);

    void dumpBacktrace();
    
    String* err;
    int code;

    char* tag;
    int arity;
    Value** args;

    wchar_t* throwfile;
    int throwline;

    // Need a copy of entire backtrace. Ick.
    vector<CallStackEntry> backtrace;
};

class Value : public gc
{
public:
    /// Create an initialised value.
    Value(void* ptr, valtype ft):m_val(ptr),m_funtable(ft) {}

    valtype getType() { return m_funtable; }

  // returns a new value pointing to the same thing
    Value* clone() {
      return new Value(m_val,m_funtable);
    }

    /// Get an integer out. Assumes you've checked/know.
    kint getInt() { return (kint)m_val; };
    /// Get a real number out
    double getReal() { 
// FIXME: I don't know if turning this check off is 'safe'. It doesn't
// throw an exception after all. Maybe it should?
#ifndef NOCHECK
	if (m_funtable==KVT_REAL)
	    return ((Real*)m_val)->number;
	else
	    return 0.0; // Not too much of an overhead here.
#else
	    return ((Real*)m_val)->number;
#endif	
    }

    /// Get the function pointer. Assumes you've checked/know.
    Closure* getFunc() { return (Closure*)m_val; }

    /// Get the string
    String* getString();
    /// Get the array.
    Array* getArray();
    /// Get the union.
    Union* getUnion();
    /// Get the exception.
    Exception* getExcept();

    /// Run the closure stored here.
    void runClosure(VMState *vm);

    /// Add more arguments from the stack to the closure stored here.
    Value* addClosureArgs(VMState *vm, int i);

    /// Coercions
    void int2str();
    void str2int();
    void str2real();
    void chr2str();
    void bool2str();
    void str2chr();
    void int2real();
    void real2int();
    void real2str();

    /// Get the tag out of a union
    int getTag() { return U_TAG(((Union*)m_val)); }


    /// Get the raw pointer out.
    void* getRaw() { return m_val; }


    /// Update to be an integer.
    void setInt(int i)
    {
	m_val=(void*)i;
	m_funtable = KVT_INT;
    }

    /// Update to be an integer and return a pointer
    Value* setIntReturn(int i) {
	setInt(i);
	return this;
    }

    /// Update to be a real
    void setReal(double i);

    /// Update to be a function.
    void setFunc(Closure* f);

    /// Update to be ana rray.
    void setString(String* str)
    {
	m_val=(void*)str;
	m_funtable = KVT_STRING;
    }

    void setArray(Array* a);

    /// Set a raw pointer.
    void setPtr(Value* p) {
//    cout << "Making " << this << " to " << p->m_val << endl;
//	memcpy(this,p,sizeof(Value*));
	m_val=p->m_val;
	m_funtable=p->m_funtable;
    }

    void addInt(kint x) {
        m_val = (void*)((kint)m_val+x);
    }

    void addVal(Value* v) {
	m_val=(void*)((kint)m_val+(kint)(v->m_val));
    }

    void subVal(Value* v) {
	m_val=(void*)((kint)m_val-(kint)(v->m_val));
    }

    void mulVal(Value* v) {
	m_val=(void*)((kint)m_val*(kint)(v->m_val));
    }

    void divVal(Value* v) {
	m_val=(void*)((kint)m_val/(kint)(v->m_val));
    }

    /// Set a value from the projection of the top stack item
    void project(VMState* vm, int i);
    void fastproject(Value* topitem, int i) {
	Union* top=(Union*)(topitem->m_val);
	Value* arg=top->args[i];
	m_val=arg->m_val;
	m_funtable = arg->m_funtable;
    }
    void exceptionproject(Value* topitem, int i) {
	Exception* top=(Exception*)(topitem->m_val);
	Value* arg=top->args[i];
	m_val=arg->m_val;
	m_funtable = arg->m_funtable;
    }

    void fastproject(VMState* vm, int i);

    /// Lookup a value in an array --- resize/create if necessary.
    Value* lookup(VMState* vm, int i);
    /// Get the length of an array
    int length();

    /// Get the function table out
    valtype getFunTable() { return m_funtable; }
    
private:
    void* m_val;
    valtype m_funtable;
};

// Get the actual string, considering offset
#define M_STR (m_str+m_offset)
// Get the actual length available, considering offset
#define M_ALLOC (m_alloc-m_offset)
// Prepared to waste 8k before reallocating a string (this is quite a lot...)
#define OFFSET_LIMIT 8192

class String : public gc
{
public:
    String(const wchar_t* val);
    String(wchar_t val);
    String(int len);

    wchar_t* getVal() { return M_STR; }

    int length() { return wcslen(M_STR); }
    wchar_t getIndex(int i) { return M_STR[i]; }
    void setIndex(int i,wchar_t c) { M_STR[i]=c; }
    void append(String* s);
    void append(wchar_t c);
    void append(const wchar_t* s);
    bool eq(String* s);
    bool eq(wchar_t* s);
    int cmp(String* s);
    // Increase the offset. If it goes over OFFSET_LIMIT, reallocate
    void offset(int x);
    // Chop last x characters off.
    void chop(int x);
    int hash();

  // debug/marshalling purposes
    int space() { return M_ALLOC; }

private:
    wchar_t* m_str;
    int m_alloc;
    int m_offset; // How much of the start of the string to ignore
                  // (Optimisation if we're just chopping off the start of
                  //  the string)
};

#endif

==> kaya_0.2.7.orig.tar.gz: kaya-0.2.7/rts/Heap.h <==
#ifndef _HEAP_H // -*-C++-*-
#define _HEAP_H

/**
    Kaya run-time system
    Copyright (C) 2004, 2005 Edwin Brady

    This file is distributed under the terms of the GNU Lesser General
    Public Licence. See COPYING for licence.
*/

using namespace std;

#ifndef WIN32
#define SWPRINTF swprintf
#else
#define SWPRINTF snwprintf
#endif
#ifndef NOCHECK
#define GC_DEBUG
#endif
#ifndef WIN32
#include <pthread.h>
#define GC_THREADS
#else
#define GC_WIN32_THREADS
#endif
#include <gc/gc_cpp.h>

//#include <gc/gc_allocator.h>
#include <vector>
#include <cstdlib>
#include "sizes.h"
#include "ValueFuns.h"

class VMState;

typedef void(*func)(VMState*);

class Value;
class String;
class Array;
class Exception;
class Union;
class Closure;

#define KMALLOC(x) GC_MALLOC(x)
#define KMALLOC_BLOCK(x) GC_MALLOC_ATOMIC(x)
#define KMALLOC_ROOT(x) GC_MALLOC_UNCOLLECTABLE(x)
#define KREALLOC(x) GC_REALLOC(x)

#define NEWVAL (new Value(NULL,KVT_INT))
//(Value*)(KMALLOC(sizeof(Value))))
#define MKVAL(ptr,ft) new Value(ptr,ft)
//NEWVAL->setPtr(ptr,ft)

#define NEWREAL(n) new Real(n)
//(((Real*)(KMALLOC(sizeof(Real))))->setNum(n))

// I really shouldn't call this Real! Float/Double better.
class Real
{
public:
    void *operator new(size_t num_bytes) {
	return (KMALLOC(num_bytes));
    }

    Real(double n):number(n) {}
//    Real* setNum(double n) { number = n; return this; }
    double number;
};

// cached values
extern Value *zero, *one, *minusone;

extern "C" {
  wchar_t* emptyString();
}


//#define U_TAG(x) ((x->tag_arity) >> 16)
#define U_ARITY(x) x->arity
//#define THIS_TAG (tag_arity >> 16)
#define THIS_ARITY arity

class Union : public gc_cleanup
{
public:
    // If getargs is true, get arguments from the stack in vm.
    Union(VMState* vm, unsigned arity, bool getargs = true);
    ~Union() { if (args != NULL) { GC_FREE(args); } }

    bool eq(Union* x, map<Value*,Value*>& done);
    int cmp(Union* x);
    unsigned memusage();
    Union* copy(map<Value*,Value*>& done);

    unsigned arity; 
    Value** args;
};


#ifdef WIN32
/* No, still doesn't work. Error is:
d:/MINgw/include/c++/3.2.3/bits/stl_construct.h: In function `void 
   std::_Construct(_T1*, const _T2&) [with _T1 = CallStackEntry, _T2 = 
   CallStackEntry]':
Heap.cc:494:   instantiated from `void std::vector<_Tp, _Alloc>::push_back(const _Tp&) [with _Tp = CallStackEntry, _Alloc = std::allocator<CallStackEntry>]'
Heap.cc:455:   instantiated from here
d:/MINgw/include/c++/3.2.3/bits/stl_construct.h:78: no matching function for 
   call to `CallStackEntry::operator new(unsigned int, void*&)'
Heap.h:104: candidates are: static void* CallStackEntry::operator new(unsigned 
   int)
*/
struct CallStackEntry : public gc {
# else
class CallStackEntry {
public:
    void *operator new(size_t num_bytes) {
	return (KMALLOC(num_bytes));
    }
#endif
    const wchar_t* fn_name;
    const wchar_t* file;
    int line;
    const wchar_t* call_file;
    int call_line;
};

class Exception
{
public:
    void *operator new(size_t num_bytes) {
	return (KMALLOC(num_bytes));
    }

    Exception(VMState* vm);
    Exception(VMState* vm, const char* tag, int arity);
    void show();
    bool eq(Exception* x);
    int cmp(Exception* x);

    void dumpBacktrace();
    
    String* err;
    int code;

    const char* tag;
    int arity;
    Value** args;

    wchar_t* throwfile;
    int throwline;

    // Need a copy of entire backtrace. Ick.
    vector<CallStackEntry> backtrace;
};

class Value
{
public:
    /// Create an initialised value.
    Value(void* ptr, valtype ft):m_val(ptr),m_funtable(ft) {}

    void *operator new(size_t num_bytes) {
	return (KMALLOC(num_bytes));
    }

    valtype getType() { return m_funtable; }

  // returns a new value pointing to the same thing
    Value* clone() {
	return MKVAL(m_val,m_funtable);
    }

    /// Get an integer out. Assumes you've checked/know.
    kint getInt() { return (kint)m_val; };
    /// Get a real number out
    double getReal() { 
// FIXME: I don't know if turning this check off is 'safe'. It doesn't
// throw an exception after all. Maybe it should?
#ifndef NOCHECK
	if (m_funtable==KVT_REAL)
	    return ((Real*)m_val)->number;
	else
	    return 0.0; // Not too much of an overhead here.
#else
	    return ((Real*)m_val)->number;
#endif	
    }

    /// Get the function pointer. Assumes you've checked/know.
    Closure* getFunc() { return (Closure*)m_val; }

    /// Get the string
    String* getString();
    /// Get the array.
    Array* getArray();
    /// Get the union.
    Union* getUnion();
    /// Get the exception.
    Exception* getExcept();

    /// Run the closure stored here.
    void runClosure(VMState *vm);

    /// Add more arguments from the stack to the closure stored here.
    Value* addClosureArgs(VMState *vm, int i);

    /// Coercions
    void int2str();
    void str2int();
    void str2real();
    void chr2str();
    void bool2str();
    void str2chr();
    void int2real();
    void real2int();
    void real2str();

    /// Get the tag out of a union
    int getTag() { return m_funtable-KVT_UNION; }


    /// Get the raw pointer out.
    void* getRaw() { return m_val; }


    /// Update to be an integer.
    void setInt(kint i)
    {
	m_val=(void*)i;
	m_funtable = KVT_INT;
    }

    /// Update to be an integer and return a pointer
    Value* setIntReturn(kint i) {
	setInt(i);
	return this;
    }

    /// Update to be a real
    void setReal(double i);

    /// Update to be a function.
    void setFunc(Closure* f);

    /// Update to be ana rray.
    void setString(String* str)
    {
	m_val=(void*)str;
	m_funtable = KVT_STRING;
    }

    void setArray(Array* a);

    /// Set a raw pointer.
    void setPtr(Value* p) {
//    cout << "Making " << this << " to " << p->m_val << endl;
//	memcpy(this,p,sizeof(Value*));
	m_val=p->m_val;
	m_funtable=p->m_funtable;
    }

    Value* setPtr(void* raw, valtype ft) {
	m_val = raw;
	m_funtable = ft;
	return this;
    }

    void addInt(kint x) {
        m_val = (void*)((kint)m_val+x);
    }

    void addVal(Value* v) {
	m_val=(void*)((kint)m_val+(kint)(v->m_val));
    }

    void subVal(Value* v) {
	m_val=(void*)((kint)m_val-(kint)(v->m_val));
    }

    void mulVal(Value* v) {
	m_val=(void*)((kint)m_val*(kint)(v->m_val));
    }

    void divVal(Value* v) {
	m_val=(void*)((kint)m_val/(kint)(v->m_val));
    }

    /// Set a value from the projection of the top stack item
    void project(VMState* vm, int i);
    void fastproject(Value* topitem, int i) {
	Union* top=(Union*)(topitem->m_val);
	Value* arg=top->args[i];
	m_val=arg->m_val;
	m_funtable = arg->m_funtable;
    }
    void exceptionproject(Value* topitem, int i) {
	Exception* top=(Exception*)(topitem->m_val);
	Value* arg=top->args[i];
	m_val=arg->m_val;
	m_funtable = arg->m_funtable;
    }

    void fastproject(VMState* vm, int i);

    /// Lookup a value in an array --- resize/create if necessary.
    Value* lookup(VMState* vm, kint i);
    /// Get the length of an array
    kint length();

    /// Get the function table out
    valtype getFunTable() { return m_funtable; }
    
private:
    void* m_val;
    valtype m_funtable;
};

// Get the actual string, considering offset
#define M_STR (m_str+m_offset)
// Get the actual length available, considering offset
#define M_ALLOC (m_alloc-m_offset)
// Prepared to waste 8k before reallocating a string (this is quite a lot...)
#define OFFSET_LIMIT 8192

#define NEWSTRING(x) new String(x)
//(((String*)(KMALLOC(sizeof(String))))->setVal(x))

class String
{
public:
    String(const wchar_t* val);
    String(wchar_t val);
    String(kint len);

    void *operator new(size_t num_bytes) {
	return (KMALLOC(num_bytes));
    }

    String* setVal(const wchar_t* val);
    String* setVal(wchar_t val);
    String* setVal(kint len);

    wchar_t* getVal() { return M_STR; }

    kint length() {
        return m_len; 
    }
    wchar_t getIndex(kint i) { return M_STR[i]; }
    void setIndex(kint i,wchar_t c) { M_STR[i]=c; }
    void append(String* s);
    void append(wchar_t c);
    void append(const wchar_t* s);
    bool eq(String* s);
    bool eq(const wchar_t* s);
    int cmp(String* s);
    // Increase the offset. If it goes over OFFSET_LIMIT, reallocate
    void offset(kint x);
    // Chop last x characters off.
    void chop(kint x);
  //    int hash();

  // debug/marshalling purposes
    int space() { return M_ALLOC; }

private:
    wchar_t* m_str;
    kint m_alloc;
    kint m_offset; // How much of the start of the string to ignore
                  // (Optimisation if we're just chopping off the start of
                  //  the string)
    kint m_len;
};

#endif
